"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.initializeInboxListener = void 0;
const pg_1 = require("pg");
const utils_1 = require("../common/utils");
const logical_replication_listener_1 = require("../replication/logical-replication-listener");
const inbox_message_storage_1 = require("./inbox-message-storage");
/**
 * Initialize the listener to watch for inbox table inserts.
 * @param config The configuration object with required values to connect to the WAL.
 * @param messageHandlers A list of message handlers to handle the inbox messages.
 * @param logger A logger instance for logging trace up to error logs
 * @param concurrencyController A controller that ensures specific concurrency guarantees. Defaults to the createMutexConcurrencyController.
 * @returns Functions for a clean shutdown.
 */
const initializeInboxListener = (config, messageHandlers, logger, concurrencyController) => {
    const pool = createPgPool(config, logger);
    const messageHandlerDict = getMessageHandlerDict(messageHandlers);
    const messageHandler = createMessageHandler(messageHandlerDict, pool, config, logger);
    const errorResolver = createErrorResolver(messageHandlerDict, pool, config, logger);
    const [shutdown] = (0, logical_replication_listener_1.createLogicalReplicationListener)(config, messageHandler, errorResolver, concurrencyController, logger);
    return [
        () => __awaiter(void 0, void 0, void 0, function* () {
            pool.removeAllListeners();
            try {
                yield Promise.all([pool.end(), shutdown()]);
            }
            catch (e) {
                logger.error(e, 'Inbox listener shutdown error');
            }
        }),
    ];
};
exports.initializeInboxListener = initializeInboxListener;
const createPgPool = (config, logger) => {
    const pool = new pg_1.Pool(config.pgConfig);
    pool.on('error', (err) => {
        logger.error(err, 'PostgreSQL pool error');
    });
    return pool;
};
/**
 * Executes the inbox verification, the actual message handler, and marks the
 * inbox message as processed in one transaction.
 */
const createMessageHandler = (messageHandlers, pool, config, logger) => {
    return (message) => __awaiter(void 0, void 0, void 0, function* () {
        yield (0, utils_1.executeTransaction)(pool, (client) => __awaiter(void 0, void 0, void 0, function* () {
            const result = yield (0, inbox_message_storage_1.verifyInbox)(message, client, config);
            if (result === true) {
                const handler = messageHandlers[getMessageHandlerKey(message)];
                if (handler) {
                    yield handler.handle(message, client);
                }
                yield (0, inbox_message_storage_1.ackInbox)(message, client, config);
            }
            else {
                logger.warn(message, `Received inbox message cannot be processed: ${result}`);
            }
        }), logger);
    });
};
const getMessageHandlerKey = (h) => `${h.aggregateType}@${h.messageType}`;
const getMessageHandlerDict = (messageHandlers) => {
    const handlers = {};
    for (const handler of messageHandlers) {
        const key = getMessageHandlerKey(handler);
        if (key in handlers) {
            throw new Error(`Only one message handler can handle one aggregate and message type. Multiple message handlers try to handle the aggregate type "${handler.aggregateType}" with the message type "${handler.messageType}".`);
        }
        handlers[key] = handler;
    }
    if (Object.keys(handlers).length === 0) {
        throw new Error('At least one message handler must be provided.');
    }
    return handlers;
};
/**
 * Increase the attempts counter of the message and retry it later.
 */
const createErrorResolver = (messageHandlers, pool, config, logger) => {
    /**
     * An error handler that will increase the inbox attempts counter on transient
     * errors or sets the attempts counter to max attempts for permanent errors.
     * The message handler `handleError` is called to allow for your code to
     * decide if the error is a transient or a permanent one.
     * @param message: the InboxMessage that failed to be processed
     * @param error: the error that was thrown while processing the message
     */
    return (message, error) => __awaiter(void 0, void 0, void 0, function* () {
        const handler = messageHandlers[getMessageHandlerKey(message)];
        try {
            return yield (0, utils_1.executeTransaction)(pool, (client) => __awaiter(void 0, void 0, void 0, function* () {
                let errorType = 'transient_error';
                const maxAttempts = (0, inbox_message_storage_1.getMaxAttempts)(config.settings.maxAttempts);
                if (handler === null || handler === void 0 ? void 0 : handler.handleError) {
                    errorType = yield handler.handleError(error, message, client, {
                        current: message.attempts + 1,
                        max: maxAttempts,
                    });
                }
                let attempts;
                if (errorType === 'permanent_error' ||
                    message.attempts + 1 >= maxAttempts) {
                    attempts = maxAttempts;
                    logger.error(error, `Giving up processing the message with id ${message.id}.`);
                }
                else {
                    logger.warn(error, `An error ocurred while processing the message with id ${message.id}.`);
                }
                yield (0, inbox_message_storage_1.nackInbox)(message, client, config, attempts);
                return errorType;
            }), logger);
        }
        catch (error) {
            logger.error(Object.assign(Object.assign({}, message), { err: error }), handler
                ? 'The error handling of the message failed. Please make sure that your error handling code does not throw an error!'
                : 'The error handling of the message failed.');
            return 'transient_error';
        }
    });
};
//# sourceMappingURL=inbox-listener.js.map